import os

from .check import HealthCheck, create_url_health_check
from .registry import HealthCheckRegistry


def add_built_in_health_checks(registry: HealthCheckRegistry) -> None:
    engine_url = (
        os.getenv("PROCESSCUBE_ENGINE_URL", "http://localhost:56000").strip("/")
        + "/atlas_engine/api/v1/info"
    )
    authority_url = (
        os.getenv("PROCESSCUBE_AUTHORITY_URL", "http://localhost:56020").strip("/")
        + "/.well-known/openid-configuration"
    )

    registry.register(
        HealthCheck(
            create_url_health_check(engine_url),
            service_name="ProcessCube Engine",
            tags=["core", "backend"],
            comments=["Checks if the ProcessCube Engine is reachable"],
        )
    )
    registry.register(
        HealthCheck(
            create_url_health_check(authority_url),
            service_name="ProcessCube Authority",
            tags=["core", "auth"],
            comments=["Checks if the ProcessCube Authority is reachable"],
        )
    )



from datetime import timedelta

from pydantic import BaseModel, Field, field_serializer


class HealthCheckEntityModel(BaseModel):
    service: str
    status: str = "healthy"
    time_taken: timedelta = Field(default=timedelta())
    tags: list[str] = Field(default_factory=list)
    comments: list[str] = Field(default_factory=list)

    @field_serializer("time_taken")
    def serialize_time_taken(self, time: timedelta) -> str:
        return str(time)


class HealthCheckModel(BaseModel):
    status: str = "healthy"
    total_time_taken: timedelta = Field(default=timedelta())
    entities: list[HealthCheckEntityModel] = Field(default_factory=list)

    @field_serializer("total_time_taken")
    def serialize_total_time_taken(self, time: timedelta) -> str:
        return str(time)


class LivezResponse(BaseModel):
    status: str = Field(default="alive", description="Liveness status")


class HealthConditionInfo(BaseModel):
    service_name: str = Field(..., description="Name of the service")
    tags: list[str] = Field(default_factory=list, description="Tags for categorization")
    comments: list[str] = Field(default_factory=list, description="Comments describing the check")

class HealthCheckResult(BaseModel):
    healthy: bool = Field(..., description="Indicates if the health check passed")
    time_taken: timedelta = Field(..., description="Time taken to perform the health check")

    @field_serializer("time_taken")
    def serialize_time_taken(self, time: timedelta) -> str:
        return str(time)



import asyncio
from datetime import timedelta
from time import perf_counter

from fastapi.responses import JSONResponse

from .check import HealthCheck
from .models import (
    HealthCheckEntityModel,
    HealthCheckModel,
)
from .registry import HealthCheckRegistry


async def _run_health_checks(checks: list[HealthCheck]) -> HealthCheckModel:
    total_start = perf_counter()
    results = await asyncio.gather(*(check.run() for check in checks))
    total_elapsed = perf_counter() - total_start

    entities = []
    all_healthy = True
    for check, result in zip(checks, results):
        if not result.healthy:
            all_healthy = False
        entity = HealthCheckEntityModel(
            service=check.service_name,
            status="healthy" if result.healthy else "unhealthy",
            time_taken=result.time_taken,
            tags=check.tags,
            comments=check.comments,
        )
        entities.append(entity)

    return HealthCheckModel(
        status="healthy" if all_healthy else "unhealthy",
        total_time_taken=timedelta(seconds=total_elapsed),
        entities=entities,
    )


async def health_check(registry: HealthCheckRegistry) -> HealthCheckModel:
    health_result = await _run_health_checks(registry.get_all())
    status_code = 200 if health_result.status == "healthy" else 503
    return JSONResponse(content=health_result.model_dump(), status_code=status_code)



from collections.abc import Awaitable, Callable
from datetime import timedelta
from time import perf_counter
import inspect

import httpx

from .models import HealthCheckResult


type HealthCheckCondition = Callable[[], bool | Awaitable[bool]]

class HealthCheck:
    def __init__(
        self,
        condition: HealthCheckCondition,
        service_name: str,
        tags: list[str] | None = None,
        comments: list[str] | None = None,
    ):
        self.condition = condition
        self.service_name = service_name
        self.tags = tags or []
        self.comments = comments or []

    async def run(self) -> HealthCheckResult:
        start = perf_counter()
        try:
            result = self.condition()
            if inspect.isawaitable(result):
                result = await result
            healthy = bool(result)
        except Exception:
            healthy = False
        elapsed = perf_counter() - start
        return HealthCheckResult(healthy=healthy, time_taken=timedelta(seconds=elapsed))


def create_url_health_check(url: str, timeout: float = 5.0) -> HealthCheckCondition:
    async def check():
        async with httpx.AsyncClient() as client:
            response = await client.get(url, timeout=timeout)
            return True if response.status_code == 200 else False

    return check



import inspect
from typing import Callable, Any
from pydantic import BaseModel


def create_typed_handler_wrapper(handler: Callable) -> Callable[[dict], Any]:
    signature = inspect.signature(handler)
    params = signature.parameters

    if len(params) != 1:
        raise ValueError(
            f"Handler function must have exactly one parameter, got {len(params)}."
        )

    payload_parameter = next(iter(params.values()))
    if not issubclass(payload_parameter.annotation, BaseModel):
        raise ValueError(
            f"Handler parameter must be a subclass of BaseModel from pydantic, got {payload_parameter.annotation}."
        )

    ResultModel = signature.return_annotation
    if ResultModel is inspect.Signature.empty:
        raise ValueError("Handler function must have a return type annotation.")

    if not issubclass(ResultModel, BaseModel):
        raise ValueError(
            f"Handler return type must be a subclass of BaseModel from pydantic, got {ResultModel}."
        )

    PayloadModel = payload_parameter.annotation

    async def wrapper(raw_payload: dict) -> dict:
        parsed_payload = PayloadModel.model_validate(
            raw_payload, by_alias=False, by_name=True
        )
        result = handler(parsed_payload)
        if inspect.iscoroutine(result):
            result = await result

        if not isinstance(result, ResultModel):
            raise ValueError(
                f"Handler return value must be an instance of {ResultModel.__name__}, got {type(result).__name__}."
            )
        return result.model_dump(
            mode="json",
            by_alias=False,
        )

    return wrapper



from .check import HealthCheck, create_url_health_check
from .handlers import health_check
from .models import (
    HealthCheckEntityModel,
    HealthCheckModel,
    HealthConditionInfo,
    LivezResponse,
)
from .registry import HealthCheckRegistry
from .built_in import add_built_in_health_checks
from .routes import setup_health_routes

__all__ = [
    "add_built_in_health_checks",
    "create_url_health_check",
    "health_check",
    "HealthCheck",
    "HealthCheckEntityModel",
    "HealthCheckModel",
    "HealthCheckRegistry",
    "HealthConditionInfo",
    "LivezResponse",
]



from os import getenv

# Prevent nest_asyncio from patching asyncio.run() in processcube_client - it breaks uvicorn's loop_factory
import nest_asyncio
nest_asyncio.apply = lambda *args, **kwargs: None

from processcube_client.external_task import ExternalTaskClient
from processcube_client.app_info import AppInfoClient

from .identity_provider import IdentityProvider


def _determine_authority_url(engine_url: str) -> str:
    if authority := getenv("PROCESSCUBE_AUTHORITY_URL"):
        return authority

    app_info_client = AppInfoClient(engine_url)
    authority_url = app_info_client.get_authority()
    return authority_url # type: ignore


def create_external_task_client(
    engine_url: str,
    client_name: str,
    client_secret: str,
    client_scopes: str,
    max_get_oauth_access_token_retries: int,
) -> ExternalTaskClient:
    authority_url = _determine_authority_url(engine_url)

    identity_provider = IdentityProvider(
        authority_url,
        client_name,
        client_secret,
        client_scopes,
        max_get_oauth_access_token_retries,
    )
    client = ExternalTaskClient(
        engine_url,
        identity=identity_provider,
        install_signals=False,
    )

    return client
